#include "IOExpanderCommunicationModule.h"
#include "../ErrorHandling.h"

#include <QEventLoop>
#include <QMap>
#include <QTextStream>

static QTextStream cout(stdout);


CIOExpanderCommunicationModule::CIOExpanderCommunicationModule()
:CUSBCommunicationModule(QString("USB IO Expander"))
{
	m_configData.frequency = 0;
	m_configData.spsr = 0;
	m_configData.spcr = 0;

}

void CIOExpanderCommunicationModule::ProcessReceivedData(CUSBCommand usbCmd)
{
	QList<quint8> answerData = usbCmd.Answer();
	quint16	bytesReceived = answerData.size();
	QList<quint8> cfgData = usbCmd.getCfgData();
	char cmd = usbCmd.Cmd();

	switch(cmd)
	{
		case CMD_IN:
		{
			quint16 state = (static_cast<quint16>(answerData.takeFirst()) << 8) | static_cast<quint16>(answerData.takeFirst());
			emit NewInputReceived(state);
		}break;
		case CMD_LOGIC_OP:
		case CMD_SAMPLE:
		{
			QBitArray	p[16];

			for(int i = 0; i < 16; i++)
				p[i].resize( bytesReceived / 2 );

			for(int n = 0; n < bytesReceived / 2;n++)
			{
				quint16 s = (static_cast<quint16>(answerData.takeFirst()) << 8) | static_cast<quint16>(answerData.takeFirst());

				for(int i = 0; i < 16; i++)
				{
					bool v = (s & (1<<i)) ? 1 : 0;
					p[i].setBit(n,v);
				}
			}
			emit 	NewSamplesReceived(p);
		}break;
		case CMD_A2D:
		{
			quint8 c = 0;
			quint8 channels = cfgData.first();
			for(int n = 0; n < bytesReceived;n++)
			{
				for(; c < 4;c++)
				{
					if( channels & (1<<c) )
						break;
				}
				//c is the channel
				emit NewADCValueReceived(answerData.takeFirst(), c);
			}
		}break;
		case CMD_SPI_TRANSFER:
		{
			if(bytesReceived != 1)
				break;

			quint8 d = answerData.takeFirst();
			emit NewSPIDataReceived(d);
		}break;
		case CMD_SOFTWARE_VER:
		{
			QString sw;
			sw += QString(answerData.takeFirst());
			sw += QString(answerData.takeFirst());
			sw += QString(answerData.takeFirst());
			sw += QString(answerData.takeFirst());
			emit NewSWVersion(sw);
			SHOW_INFO(tr("USB IO Expander SW: ") + sw);
		}break;
		default:
		{
			SHOW_ERROR( tr("Ungltige Daten empfangen") );
		}break;
	}
}

void CIOExpanderCommunicationModule::ConnectionEstablishedCallback()
{
	RequestSW();
}

void CIOExpanderCommunicationModule::SetFrequency(quint32 f)
{
	if( f < 1 || f > 100000)
	{
		SHOW_ERROR( tr("Ungltige Frequenz"));
		return;
	}

	if(!CheckConnetion())
		return;

	if(m_configData.frequency == f)
		return;

	//calc prescaler and timer overflow value

	QMap <unsigned int,quint8>	prescaler;
	prescaler[1] = 1;
	prescaler[8] = 2;
	prescaler[64] = 3;
	prescaler[256] = 4;
	prescaler[1024] = 5;

	QMapIterator<unsigned int,quint8> i(prescaler);

	quint32 ocr = 0;

	while(i.hasNext())
	{
		i.next();
		ocr = static_cast<quint32>(static_cast<float>(s_cpuFrequency) / i.key() / f + 1.0f);
		if(ocr && ocr < 0xFFFF)
			break;
	}

	if(ocr == 0)
	{
		SHOW_ERROR( tr("Ungltiger OCR Wert"));
		return;
	}

	m_configData.frequency = f;

	//set frequency
	CUSBCommand cmd(CMD_FREQ);
	cmd << i.value();
	cmd << static_cast<quint8>((ocr >> 8) & 0xFF);
	cmd << static_cast<quint8>(ocr & 0xFF);

	SendCommand(cmd);

//	SHOW_INFO( QString("i") + QString::number(i.value() ));
//	SHOW_INFO( QString("ocr") + QString::number( ocr ));
}

void CIOExpanderCommunicationModule::SetTrigger(quint16 trigger1, quint16 trigger2)
{
	if(!CheckConnetion())
		return;

	if(m_configData.trigger1 == trigger1 && m_configData.trigger2 == trigger2)
		return;

	m_configData.trigger1 = trigger1;
	m_configData.trigger2 = trigger2;

	CUSBCommand cmd(CMD_SET_TRIGGER);
	cmd << static_cast<quint8>((trigger1 >> 8) & 0xFF);
	cmd << static_cast<quint8>(trigger1 & 0xFF);
	cmd << static_cast<quint8>((trigger2 >> 8) & 0xFF);
	cmd << static_cast<quint8>(trigger2 & 0xFF);

	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::ConfigSPI(quint8 spcr, quint8 spsr)
{
	if(!CheckConnetion())
		return;

	if(m_configData.spcr == spcr && m_configData.spsr == spsr)
		return;

	m_configData.spcr = spcr;
	m_configData.spsr = spsr;

	CUSBCommand cmd(CMD_SPI_CONFIG);
	cmd << spcr;
	cmd << spsr;
	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::RequestSW()
{
	if(!CheckConnetion())
		return;

	CUSBCommand	cmd(CMD_SOFTWARE_VER);
	cmd.setAnswerBytes(4);
	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::Reset()
{
	if(!CheckConnetion())
		return;

	SHOW_INFO(tr("USB IO Expander Reseted"));

	CUSBCommand	cmd(CMD_RESET);
	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::RequestSamples(quint8 num, quint16 frequ, quint16 trigger1, quint16 trigger2)
{
	if(!CheckConnetion())
		return;

	SetFrequency(frequ);
	SetTrigger(trigger1,trigger2);

	//ask for samples

	CUSBCommand	cmd(CMD_SAMPLE, (trigger1 | trigger2)? 1:0);

	cmd << num;
	cmd.setAnswerBytes(num * 2); //2 bytes pro sample

	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::SendPattern(QBitArray *pattern, quint16 frequ, quint16 trigger1, quint16 trigger2)
{
	if(!CheckConnetion())
		return;

	SetFrequency(frequ);
	SetTrigger(trigger1,trigger2);

	CUSBCommand cmd(CMD_PATTERN);

	cmd << pattern->size();

	//send pattern
	for(int n = 0; n < pattern->size();n++)
	{
		quint16 s = 0;
		for(int i = 0; i < 16;i++)
		{
			bool v = pattern[i].at(n);
			s |= (v<<i);
		}
		cmd << static_cast<quint8>((s >> 8) & 0xFF);
		cmd << static_cast<quint8>(s & 0xFF);
	}

	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::RequestSamplesAndSendPattern(QBitArray *pattern, quint16 frequ, quint16 trigger1, quint16 trigger2)
{
	if(!CheckConnetion())
		return;

	SetFrequency(frequ);
	SetTrigger(trigger1,trigger2);

	int num = pattern[0].size();

	CUSBCommand cmd(CMD_LOGIC_OP, (trigger1 | trigger2)? 1:0);

	cmd << num;

	//send pattern
	for(int n = 0; n < num;n++)
	{
		quint16 s = 0;
		for(int i = 0; i < 16;i++)
		{
			bool v = pattern[i].at(n);
			s |= (v<<i);
		}
		cmd << static_cast<quint8>((s >> 8) & 0xFF);
		cmd << static_cast<quint8>(s & 0xFF);
	}

	cmd.setAnswerBytes(num * 2); //2 bytes pro sample

	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::SetOutput(quint16 out)
{
	if(!CheckConnetion())
		return;

	CUSBCommand cmd(CMD_OUT);

	cmd << static_cast<quint8>((out >> 8) & 0xFF);
	cmd << static_cast<quint8>(out & 0xFF);

	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::SetRelais(quint8 out)
{
	if(!CheckConnetion())
		return;


	CUSBCommand cmd(CMD_RELAIS);

	cmd << out;

	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::PollInput()
{
	if(!CheckConnetion())
		return;

	CUSBCommand cmd(CMD_IN);
	cmd.setAnswerBytes(2);

	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::PollADC(quint8 channels)
{
	if(!CheckConnetion())
		return;

	CUSBCommand cmd(CMD_A2D);

	cmd << channels;

	int numChannels = 0;

	for(int n = 0; n < 4; n++)
	{
		if( channels & (1<<n) )
			numChannels++;
	}

	cmd.setAnswerBytes(numChannels);
	cmd.setCfgData(channels);

	SendCommand(cmd);
}

void CIOExpanderCommunicationModule::SendSPIData(quint8 spcr, quint8 spsr, QList<quint8> data)
{
	if(!CheckConnetion())
		return;

	if(data.isEmpty())
		return;

	ConfigSPI(spcr,spsr);
	quint8 d;
	foreach(d,data)
	{
		CUSBCommand cmd(CMD_SPI_TRANSFER);
		cmd << d;
		cmd.setAnswerBytes(1);
		SendCommand(cmd);
	}
}

void CIOExpanderCommunicationModule::SetServos(QList<float> values,float min,float max)
{
	CUSBCommand cmd(CMD_SET_SERVO);

	#define	SERVO_PRESCALER	8

	//tcr0 berechnen

	for(int i = 0; i < values.size();i++)
	{
		float pulselength = min + (max-min) * values[i];

		float timer_frequ = s_cpuFrequency / static_cast<float>(SERVO_PRESCALER);

		quint16 tcr = static_cast<quint16> (pulselength / 1000.0 * timer_frequ);

		cmd << static_cast<quint8>(tcr >> 8);
		cmd << static_cast<quint8>(tcr & 0xFF);
	}

	SendCommand(cmd);
}


bool CIOExpanderCommunicationModule::CheckConnetion()
{
	if( !IsConnected() )
	{
		m_configData.frequency = 0;
		m_configData.spsr = 0;
		m_configData.spcr = 0;
		SHOW_ERROR(tr("Keine Verbindung"));
		return 0;
	}
	return 1;
}


unsigned int		CIOExpanderCommunicationModule::s_cpuFrequency = 14745600L;
